package authentication.callback;

import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.HashMap;

import javax.security.auth.callback.*;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import com.sap.engine.lib.security.Base64;
import com.sap.engine.services.servlets_jsp.server.application.WebApplicationConfig;
import com.sap.tc.logging.Location;
import com.sap.tc.logging.Severity;

/**
 * Callback handler to be used by web applications.
 */
public class HttpCallbackHandler implements CallbackHandler {
	
  /**
   * The key to the error page of the web application.
   */
  public static final String FORM_ERROR_PAGE = "/form_pages/ErrorPage.jsp";
  /**
   * The key to the login page of the web application.
   */
  public static final String FORM_LOGIN_PAGE = "/form_pages/LoginPage.jsp";
   
  protected HttpServletRequest request;
  protected HttpServletResponse response;
  
  private static final String USERNAME = "j_username";
  private static final String PASSWORD = "j_password";
  private static final String J_SECURITY_CHECK = "j_security_check";
  private static final String DEFAULT_AUTH_TYPE = "basic";
  private static final String DEFAULT_POLICY_DOMAIN = "/default";

  private static Location LOCATION = Location.getLocation("authentication.framework");
  
  private boolean usernameOrPasswordRequested = false;
  private char[] password = null;
  private String userName = null;
  
  private String authType = null;
  private String loginPage = null;
  private String errorPage = null;
  private String policyDomain = null;
  
  private HashMap<String, Cookie> cookies = null;

  public HttpCallbackHandler(HttpServletRequest request, HttpServletResponse response) {
  	this.request = request;
    this.response = response;

    this.authType = DEFAULT_AUTH_TYPE;
    this.loginPage = FORM_LOGIN_PAGE;
    this.errorPage = FORM_ERROR_PAGE;
    this.policyDomain = DEFAULT_POLICY_DOMAIN;
  }
  
  public HttpCallbackHandler(HttpServletRequest request, HttpServletResponse response, String authType, String loginPage, String errorPage, String policyDomain) {
    this.request = request;
    this.response = response;
    
  	if (authType != null) {
  		this.authType = authType;
  	} else {
  		this.authType = DEFAULT_AUTH_TYPE;
  	}
  	
  	this.loginPage = getProperty(loginPage, FORM_LOGIN_PAGE);
  	this.errorPage = getProperty(errorPage, FORM_ERROR_PAGE);
  	this.policyDomain = getProperty(policyDomain, DEFAULT_POLICY_DOMAIN);
  }
  
  public HttpCallbackHandler(HttpServletRequest request, HttpServletResponse response, WebApplicationConfig applicationConfiguration) {
  	this(request, response, applicationConfiguration.getAuthType(), applicationConfiguration.getLoginPage(), applicationConfiguration.getErrorPage(), applicationConfiguration.getDomainName());
  }

  public void handle(Callback[] callbacks) throws IOException, UnsupportedCallbackException {
    for (int i = 0; i < callbacks.length; i++) {
      Callback callback = callbacks[i];

      if (LOCATION.beDebug()) {
        LOCATION.debugT("handle", "Handle " + callback);
      }

      if (callback == null) {
        continue;
      }
      
      if (callback instanceof PasswordCallback) {
        handle((PasswordCallback) callback);
        
      } else if (callback instanceof NameCallback) {
        handle((NameCallback) callback);
        
      } else if (callback instanceof CertificateCallback) {
        handle((CertificateCallback) callback);
        
      } else if (callback instanceof SSOTokenGetterCallback) {
        handle((SSOTokenGetterCallback) callback);
        
      } else if (callback instanceof SSOTokenSetterCallback) {
        handle((SSOTokenSetterCallback) callback);
        
      } else if (callback instanceof AuthenticationStatusCallback) {
      	handle((AuthenticationStatusCallback) callback);
      	
      } else {
        if (LOCATION.beDebug()) {
          LOCATION.debugT("handle", "Not supported callback: " + callback);
        }
        
        throw new UnsupportedCallbackException(callback);
      }
    }
  }

  private void handle(NameCallback nameCallback) {
    parseUserNameAndPassword();
    nameCallback.setName(this.userName);

    if (LOCATION.beDebug()) {
      LOCATION.debugT("handle", "Set name in NameCallback to " + userName);
    }
  }

  private void handle(PasswordCallback passwordCallback) {
    parseUserNameAndPassword();
    passwordCallback.setPassword(this.password);
    
    if (LOCATION.beDebug()) {
      LOCATION.debugT("handle", "Set password in PasswordCallback: " + (this.password != null));
    }
  }
  
  private void handle(CertificateCallback certificateCallback) {
    X509Certificate[] certificate = (X509Certificate[]) request.getAttribute("javax.servlet.request.X509Certificate");
    certificateCallback.setCertificates(certificate);
    
    if (LOCATION.beDebug()) {
      LOCATION.debugT("handle", "Set certificate in CertificateCallback: " + certificate);
    }
  }
  
  private void handle(SSOTokenGetterCallback getterCallback) {
    String name = getterCallback.getName();

    if (LOCATION.beDebug()) {
      LOCATION.debugT("Get cookie " + name);
    }

    HashMap allCookies = getCookies();
    Cookie cookie = (Cookie) allCookies.get(name);

    if (cookie != null) {
      getterCallback.setValue(cookie.getValue());

      if (LOCATION.beDebug()) {
      	LOCATION.debugT("Set value to " + cookie.getValue());
      }
    } else {
      getterCallback.setValue (null);

      if (LOCATION.beDebug()) {
        LOCATION.debugT("Cookie " + name + " is not found");
      }
    }
  }
  
  private void handle(SSOTokenSetterCallback setterCallback) {
    String name = setterCallback.getName();

    if (name != null) {
      Cookie cookie = new Cookie(name, (String) setterCallback.getValue());
      this.response.addCookie(cookie);

      if (LOCATION.beDebug()) {
      	LOCATION.debugT("Added cookie " + name + " with value " + setterCallback.getValue());
      }
    }
  }
  
  private void handle(AuthenticationStatusCallback statusCallback) throws IOException {
    if (response.isCommitted()) {
      if (LOCATION.beInfo()) {
        LOCATION.infoT("handle", "The response is commited. No redirect is possible.");
      }
      
      return;
    }

    if (statusCallback == AuthenticationStatusCallback.PASSED) {
      goToOriginalPage();
    } else if (statusCallback == AuthenticationStatusCallback.FAILED) {
    	goToLoginOrErrorPage();
    }
  }
  
  private void goToFormLoginPage() throws IOException {
  	String url = request.getContextPath() + loginPage;
    if (LOCATION.beDebug()) {
      LOCATION.debugT("Redirect to login page: " + url);
    }
    
  	HttpSession session = request.getSession(true);
  	
  	if (session.getAttribute("original.url") == null) {
  		session.setAttribute("original.url", request.getRequestURL().toString());
  		
  		if (LOCATION.beDebug()) {
        LOCATION.debugT("Set original url : " + request.getRequestURL());
      }
  	}
    
    sendRedirect(url);
  }
  
  private void goToFormErrorPage() throws IOException {
  	String url = request.getContextPath() + errorPage;
    if (LOCATION.beDebug()) {
      LOCATION.debugT("Forward to error page: " + url);
    }
        
    sendRedirect(url);
  }

  private void goToLoginOrErrorPage() throws IOException {
    final String methodName = "redirectToLoginOrErrorPage";

    if (LOCATION.bePath()) {
      LOCATION.entering(methodName);
    }

    try {
      if (HttpServletRequest.FORM_AUTH.equalsIgnoreCase(authType)) {
        if (this.response.isCommitted()) {
          if (LOCATION.beDebug()) {
            LOCATION.debugT(methodName, "Response is already committed");
          }
          return;
        }

        if (!usernameOrPasswordRequested) {
          if (LOCATION.beDebug()) {
            LOCATION.debugT(methodName, "Cannot forward to login/error page. No username or password has been requested by the login modules.");
          }
          return;
        }

        if (request.getRequestURI().endsWith(J_SECURITY_CHECK)) {
          goToFormErrorPage();
        } else {
          goToFormLoginPage();
        }
      } else if (HttpServletRequest.BASIC_AUTH.equalsIgnoreCase(authType)) {
        response.addHeader("WWW-Authenticate", "Basic realm=\"default\"");

        if (LOCATION.beDebug()) {
          LOCATION.debugT(methodName, "Added \"WWW-Authenticate\" header");
        }

        sendError(HttpServletResponse.SC_UNAUTHORIZED);
      } else if (HttpServletRequest.CLIENT_CERT_AUTH.equalsIgnoreCase(authType) || "CLIENT-CERT".equalsIgnoreCase(authType)) {
      	sendError(HttpServletResponse.SC_UNAUTHORIZED);
      } else if (LOCATION.beDebug()) {
        LOCATION.debugT(methodName, "Unknown authentication method " + authType);
      }
    } finally {
      if (LOCATION.bePath()) {
        LOCATION.exiting(methodName);
      }
    }
  }  
  
  private void sendRedirect(String url) throws IOException {
    if (!response.isCommitted()) {
    	String parameters = request.getQueryString();
    		
    	if ((parameters != null) && !parameters.equals("")) {
    		url = url + "?" + parameters;
    	}
      response.sendRedirect(response.encodeRedirectURL(url));

      if (LOCATION.beDebug()) {
        LOCATION.debugT("sendRedirect", "Redirected to " + url);
      }
    } else if (LOCATION.beDebug()) {
      LOCATION.debugT("sendRedirect", "Cannot send redirect because response is committed");
    }
  }

  private void sendError(int errorCode) throws IOException {
      if (!response.isCommitted()) {
        response.sendError(errorCode);

        if (LOCATION.beDebug()) {
          LOCATION.debugT("sendError", "Send error " + errorCode);
        }
      } else if (LOCATION.beDebug()) {
        LOCATION.debugT("sendError", "Cannot send error " + errorCode + " because response is already committed");
      }
  }

  private void goToOriginalPage() throws IOException {
    final String methodName = "redirectToOriginalPage";

    if (LOCATION.bePath()) {
      LOCATION.entering(methodName);
    }

    try {
    	if (request.getRequestURI().endsWith(J_SECURITY_CHECK)) {
	    	HttpSession session = request.getSession(true);
	    	String url = (String) session.getAttribute("original.url");
	    	
	    	LOCATION.infoT("original url: " + url);
	    	
	    	if ((url == null) || url.endsWith(J_SECURITY_CHECK)) {
	    		sendError(HttpServletResponse.SC_NOT_FOUND);
	    		return;
	    	}
	      	      
	    	url = getRelativeURL(url);
	    	session.removeAttribute("original.url");
	    	sendRedirect(url);
    	}
    } finally {
      if (LOCATION.bePath()) {
        LOCATION.exiting(methodName);
      }
    }
  }

  private void parseUserNameAndPassword() {
    final String methodName = "parseUserNameAndPassword";

    if (userName != null && password != null) {
      return;
    }

    if (!usernameOrPasswordRequested) {
    	usernameOrPasswordRequested = true;
      if (LOCATION.beDebug()) {
        LOCATION.debugT(methodName, "Enabled redirect to login/error page");
      }
    }

    if (HttpServletRequest.FORM_AUTH.equalsIgnoreCase(authType)) {
      initFormCredentials();
    } else if (HttpServletRequest.BASIC_AUTH.equalsIgnoreCase(authType)) {
      initBasicCredentials();
    }
  }

  private void initBasicCredentials() {
    final String methodName = "initBasicCredentials";

    String [] authHeader = parseAuthHeader(request.getHeader("Authorization"));

    if (authHeader != null) {
      this.userName = authHeader[0];

      if (LOCATION.beDebug()) {
        LOCATION.debugT(methodName, "User name in Authorization header is " + userName);
      }

      if (this.password == null) {
        this.password = authHeader[1].toCharArray();
        if (LOCATION.beDebug()) {
          LOCATION.debugT(methodName, "Get password from Authorization header");
        }
      }
    }
  }

  private void initFormCredentials() {
    final String methodName = "initFormCredentials";

    this.userName = request.getParameter(USERNAME);

    if (LOCATION.beDebug()) {
      LOCATION.debugT(methodName, "User name in the request is " + userName);
    }

    String jPassword = this.request.getParameter(PASSWORD);

    if (jPassword != null) {
      this.password = jPassword.toCharArray();

      if (LOCATION.beDebug()) {
        LOCATION.debugT(methodName, "Found password");
      }
    }
  }

  private String[] parseAuthHeader(String authorization) {
    final String methodName = "parseAuthHeader";

    if (authorization == null) {
      if (LOCATION.beDebug()) {
        LOCATION.debugT(methodName, "Provided authorization header name is null");
      }
      return null;
    }

    int startAuthDataIndex = authorization.indexOf(' ');

    if (startAuthDataIndex <= 0) {
      if (LOCATION.beDebug()) {
        LOCATION.debugT(methodName, "No method found");
      }
      return null;
    }

    String method = authorization.substring(0, startAuthDataIndex);

    if (LOCATION.beDebug()) {
      LOCATION.debugT(methodName, "Method is " + method);
    }

    if ("Basic".equals(method)) {
      String credentialsString = authorization.substring(startAuthDataIndex + 1);

      try {
      	String charsetName = request.getCharacterEncoding();
      	
      	if (charsetName == null) {
      		charsetName = "ISO-8859-1";
      	}
      	
      	byte[] decodedBytes = Base64.decode(credentialsString.getBytes());
      	
     		credentialsString = new String(decodedBytes, charsetName);
      } catch (Exception e) {
        if (LOCATION.beDebug()) {
          LOCATION.traceThrowableT(Severity.DEBUG, methodName, e);
        }
        
        return null;
      }

      String[] credentials = new String[2];
      credentials[0] = credentialsString.substring(0, credentialsString.indexOf(':'));
      credentials[1] = credentialsString.substring(credentialsString.indexOf(':') + 1);

      if (LOCATION.beDebug()) {
        LOCATION.debugT(methodName, "Username: " + credentials[0]);
      }

      return credentials;
    }

    if (LOCATION.beDebug()) {
      LOCATION.debugT(methodName, "Unsupported method: " + methodName);
    }
    
    return null;
  }
  
  private static String getRelativeURL(String url) {
  	if (url == null) {
  		return null;
  	}
  	
  	int pos = url.indexOf("/", 8);
  	
  	if (pos != -1) {
  		url = url.substring(pos);
  	}
  	
  	return url;
  }
  
  public String getSecurityPolicyDomain() {
  	return this.policyDomain;
  }
  
  public void setAuthenticationStack(String authStack) {
  	HttpSession session = request.getSession(true);
  	session.setAttribute("authentication.stack.name", authStack);
  }
  
  private String getProperty(String property, String defaultValue) {
  	String result = null;
  	
  	if (property == null) {
  		result = defaultValue;
  	} else if (!property.startsWith("/")) {
  		result = "/" + property;
  	} else {
  		result = property;
  	}
  	
  	return result;
  }
  
  private HashMap getCookies() {
    if (this.cookies == null) {
      this.cookies = new HashMap<String, Cookie>();

      Cookie[] requestCookies = request.getCookies();

      if (requestCookies != null) {
        for (int i = 0; i < requestCookies.length; i++) {
          this.cookies.put(requestCookies[i].getName(), requestCookies[i]);
        }
      }
    }
    
    return this.cookies;
  }
}
