package authentication.context;

import java.security.PrivilegedExceptionAction;

import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import authentication.configuration.AppConfigurationEntryExtension;
import authentication.configuration.ControlFlag;

import com.sap.tc.logging.Location;
import com.sap.tc.logging.Severity;

/**
 * PrivilegedExceptionAction for calling login modules.
 */
public class AuthStackProcessAction implements PrivilegedExceptionAction {

  private static final Location LOCATION = Location.getLocation("authentication.framework");
    
  private AppConfigurationEntryExtension[] configuration = null;
  private LoginModule[] modules = null;
  private byte type = -1;
  String authStack = null;

  public AuthStackProcessAction() {
  }

  public Object run() throws LoginException {
    boolean processed = false;
    boolean localProcessed = false;
    boolean onlyFinal = false;
    int i = 0;
    LoginException exception = null;
    LoginException weakException = null;

    LOCATION.errorT("modules.length " + modules.length);
    for (; i < modules.length; i++) {
      if ((!onlyFinal || (configuration[i].getControlFlagExtension() == ControlFlag.CLOSING))) {
	      localProcessed = false;

	      try {
	        switch (type) {
	          case CustomLoginContext.OPERATION_LOGIN: {            
              localProcessed = modules[i].login();
              break;
	          }
	          case CustomLoginContext.OPERATION_ABORT: {
              localProcessed = modules[i].abort();
              break;
	          }
	          case CustomLoginContext.OPERATION_COMMIT: {
              localProcessed = modules[i].commit();
              break;
	          }
	          case CustomLoginContext.OPERATION_LOGOUT: {
              localProcessed = modules[i].logout();
              break;
	          }
	        }
	
	        if (localProcessed) {
	          weakException = null;
	          processed = true;
	
	          if ((type == CustomLoginContext.OPERATION_LOGIN) || (type == CustomLoginContext.OPERATION_COMMIT) && (configuration[i].getControlFlagExtension() == ControlFlag.SUFFICIENT)) {
	          	onlyFinal = true;
	          }
	        }
	      } catch (LoginException e) {	

	        if (configuration[i].getControlFlagExtension() == ControlFlag.REQUISITE) {
	          if ((type == CustomLoginContext.OPERATION_ABORT) || (type == CustomLoginContext.OPERATION_LOGOUT)) {
	            if (exception == null) {
	              exception = e;
	            }
	          } else {
	            throw e;
	          }
	        } else if ((configuration[i].getControlFlagExtension() == ControlFlag.REQUIRED) || (configuration[i].getControlFlagExtension() == ControlFlag.CLOSING)) {
	          if (exception == null) {
	            exception = e;
	          }
	        } else {
	          if (weakException == null) {
	            weakException = e;
	          }
	        }
	      } catch (Throwable e) {
	        LOCATION.traceThrowableT(Severity.ERROR, "Login module {0} from authentication stack {1} errors while authenticating the caller. Most probably the authentication stack is not set up correctly.", new Object[] {modules[i].getClass().getName(), this.authStack}, e);
	        throw new SecurityException("Error during login: " + e.toString());
	      }
	    }
    }

    if (exception != null) {     
      if (exception instanceof LoginException) {
        throw (LoginException) exception;
      } else {
        throw new LoginException(exception.toString());
      }
    }

    if (!processed && (weakException != null)) {
      if (weakException instanceof LoginException) {
        throw weakException;
      } else {
        throw new LoginException(weakException.toString());
      }
    }

    if (!processed) {
      if (LOCATION.beDebug()) {
        LOCATION.debugT("Unsuccessful login: no login module succeeded. The size of the used authentication stack {0} is {1}.", new Object[] {this.authStack, new Integer(modules.length)});
      }
      throw new LoginException("No login module succeeded.");
    }
    
    return null;
  }

  protected AuthStackProcessAction getAction(AppConfigurationEntryExtension[] configuration, LoginModule[] modules, byte type, String authStack) {
    this.authStack = authStack;
    this.configuration = configuration;
    this.modules = modules;
    this.type = type;
    return this;
  }

}
