package authentication.context;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.security.AccessController;
import java.security.Principal;
import java.security.PrivilegedActionException;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import authentication.callback.AuthenticationStatusCallback;
import authentication.callback.HttpCallbackHandler;
import authentication.configuration.AppConfigurationEntryExtension;
import authentication.principal.UserIdPrincipal;

import com.sap.engine.frame.core.thread.ThreadContext;
import com.sap.engine.services.security.SecurityServerFrame;
import com.sap.engine.services.security.login.CreateSessionAction;
import com.sap.engine.services.security.login.LoginContextConfiguration;
import com.sap.engine.services.security.login.SecurityContext;
import com.sap.engine.services.security.login.SecuritySession;
import com.sap.engine.services.security.login.SubjectWrapper;
import com.sap.engine.services.security.server.AuthenticationContextImpl;
import com.sap.tc.logging.Location;
import com.sap.tc.logging.Severity;

public class CustomLoginContext extends LoginContext {
	
  protected final static byte OPERATION_LOGIN = 0;
  protected final static byte OPERATION_ABORT = 1;
  protected final static byte OPERATION_COMMIT = 2;
  protected final static byte OPERATION_LOGOUT = 3;

  private static final Location LOCATION = Location.getLocation("authentication.framework");

  protected Subject subject = null;
  protected Principal principal = null;
  protected CallbackHandler callbackHandler = null;

  private static int threadContextId = -1;

  private boolean isSubjectProvided = false;
  private AppConfigurationEntryExtension[] configuration = new AppConfigurationEntryExtension[0];
  private LoginModule[] modules = null;
  private AuthStackProcessAction processAction = new AuthStackProcessAction();
  private String authenticationStack = null;
  private boolean isLoginSucceeded = false;
  
  private Class[] moduleClasses = null;

  CustomLoginContext(String authenticationStack, Subject subject, CallbackHandler callbackHandler) throws LoginException {
    super(LoginContextConfiguration.EMPTY_STACK);
    this.authenticationStack = authenticationStack;
    this.subject = subject;
    this.callbackHandler = callbackHandler;
    this.isSubjectProvided = (subject != null);
    LOCATION.infoT("Login context name: " + authenticationStack);
    
    init();
  }
  
  public void login() throws LoginException {
    if (LOCATION.beInfo()) {
      LOCATION.infoT("Authentication requested.");
    }
    
  	if (callbackHandler instanceof HttpCallbackHandler) {
  		((HttpCallbackHandler) callbackHandler).setAuthenticationStack(authenticationStack);
  	}
    
    isLoginSucceeded = false;

    // TODO check wheter we need to set the session into the thread context.
    ThreadContext currentThreadContext = SecurityServerFrame.threadContext.getThreadContext();
    SecurityContext context = null;

    if (currentThreadContext != null) {
      if (threadContextId == -1) {
        threadContextId = currentThreadContext.getContextObjectId(SecurityContext.NAME);
      }
      
      try {
        com.sap.engine.session.usr.UserContext.getAccessor().apply();
      } catch (Exception e) {
      	LOCATION.traceThrowableT(Severity.ERROR, "", e);
      }
      
      context = (SecurityContext) currentThreadContext.getContextObject(threadContextId);
    }
    
    initializeLoginContext();

    if (!isSubjectProvided) {
      subject = new Subject();
    }

    SubjectWrapper subjectWrapper = null;
        
    try {    
    	LOCATION.errorT("Authentication stack during login: " + authenticationStack);
      AccessController.doPrivileged(processAction.getAction(configuration, modules, OPERATION_LOGIN, authenticationStack));
      AccessController.doPrivileged(processAction.getAction(configuration, modules, OPERATION_COMMIT, authenticationStack));
      
      subjectWrapper = new SubjectWrapper(subject, getPrincipal());
      isLoginSucceeded = true;
      notifyAuthStatus(AuthenticationStatusCallback.PASSED);
    } catch (Exception e) {
    	try {
        AccessController.doPrivileged(processAction.getAction(configuration, modules, OPERATION_ABORT, authenticationStack));
      } catch (PrivilegedActionException ee) {
      	LOCATION.traceThrowableT(Severity.DEBUG, "", ee);
      } 
      
      if (!isSubjectProvided) {
        subject = null;
      }


      Exception exception = null;
      
      if (e instanceof PrivilegedActionException) {
        exception = ((PrivilegedActionException) e).getException();
      } else {
        exception = e;
      }
      
      LOCATION.traceThrowableT(Severity.ERROR, "", exception);

      LoginException loginException = null;

      if (exception instanceof LoginException) {
        loginException = (LoginException) exception;
      } else {
        loginException = new LoginException("Authentication failed.");
      }

      notifyAuthStatus(AuthenticationStatusCallback.FAILED);
      throw loginException;
    } catch (Error e) {
      throw e;
    }

    if ((context != null)) {      
    	SecuritySession session = (SecuritySession) new CreateSessionAction(subjectWrapper, authenticationStack, AuthenticationContextImpl.getSessionPool()).run();
    	
    	if (callbackHandler instanceof HttpCallbackHandler) {    		
        String policyDomain = ((HttpCallbackHandler) callbackHandler).getSecurityPolicyDomain();
        LOCATION.infoT("Policy domain retrieved from the callback handler is " + policyDomain);
        session.addSecurityPolicyDomain(policyDomain);
      }
    	
    	context.setSession(session);
    }

    LOCATION.infoT("Authentication is successful.");
  }

  public void logout() throws LoginException {
      if (subject == null) {
        throw new LoginException("Call login before logout.");
      }

      try {
        AccessController.doPrivileged(processAction.getAction(configuration, modules, OPERATION_LOGOUT, authenticationStack));
      } catch (PrivilegedActionException e) {
        if (!isSubjectProvided) {
          subject = null;
        }

        Exception exception = e.getException();

	      if (!(exception instanceof LoginException)) {
	      	exception = new LoginException("Logout failed.");
	      }
	      
	      throw (LoginException) exception;
      }  
  }

  private void notifyAuthStatus(AuthenticationStatusCallback cbk) {
    try {
      if (callbackHandler != null) {
        this.callbackHandler.handle(new Callback[] {cbk});
      }
    } catch (Exception e) {
      if (LOCATION.beDebug()) {
        LOCATION.traceThrowableT(Severity.DEBUG, "Failed to send authentication status callback.", e);
      }
    }
  }
  
  public Subject getSubject() {
    if (isLoginSucceeded && isSubjectProvided) {
      return subject;
    }
    
    return null;
  }

  private Principal getPrincipal() {
    if (this.principal != null) {
      return this.principal;
    } else if (this.subject != null) {
      Set set = this.subject.getPrincipals(UserIdPrincipal.class);
      Iterator iterator = set.iterator();
      
      if (iterator.hasNext()) {
      	this.principal = (Principal) iterator.next();
      }

      if (this.principal == null) {
      	set = subject.getPrincipals();
        iterator = set.iterator();
        
        if (iterator.hasNext()) {
        	this.principal = (Principal) iterator.next();
        }
      }
    }

    return this.principal;
  }
  
  private void init() throws LoginException {
    if (moduleClasses == null || moduleClasses.length == 0) {
    	initializeAuthenticationStack();

      this.moduleClasses = new Class[configuration.length];
      ClassLoader classloader = getClass().getClassLoader();
      ClassLoader applicationLoader = Thread.currentThread().getContextClassLoader();

      for (int i = 0; i < moduleClasses.length; i++) {
        if (applicationLoader != null) {
          try {
            moduleClasses[i] = applicationLoader.loadClass(configuration[i].getLoginModuleName());
          } catch (ClassNotFoundException t) {
          	LOCATION.traceThrowableT(Severity.WARNING, "Cannot load login module class {0}.", new Object[] {configuration[i].getLoginModuleName()}, t);
            moduleClasses[i] = null;
          }
        }
        if (moduleClasses[i] == null) {
          if (classloader != null) {
            try {
              moduleClasses[i] = classloader.loadClass(configuration[i].getLoginModuleName());
            } catch (ClassNotFoundException t) {
            	LOCATION.traceThrowableT(Severity.WARNING, "Cannot load login module class {0}.", new Object[] {configuration[i].getLoginModuleName()}, t);
              moduleClasses[i] = null;
              
              throw new LoginException(t.toString());
            }
          } else {
            try {
              moduleClasses[i] = Class.forName(configuration[i].getLoginModuleName());
            } catch (ClassNotFoundException t) {
            	LOCATION.traceThrowableT(Severity.WARNING, "Cannot load login module class {0}.", new Object[] {configuration[i].getLoginModuleName()}, t);
              moduleClasses[i] = null;
              
              throw new LoginException(t.toString());
            }
          }
        }
      }
    }
  }

  private Map initializeLoginContext()throws LoginException {
    HashMap sharedState = new HashMap();
    this.modules = new LoginModule[moduleClasses.length];

    if (this.subject == null) {
      this.subject = new Subject();
    }
    
    for (int i = 0; i < this.modules.length; i++) {
    	try {
    		this.modules[i] = (LoginModule) moduleClasses[i].newInstance();
    	} catch (Exception e) {
    		LOCATION.traceThrowableT(Severity.ERROR, "Error during instantiation of login module " + moduleClasses[i], e);
    		this.modules[i] = null;
    		
    		throw new LoginException(e.toString());
    	}
    }

    for (int i = 0; i < moduleClasses.length; i++) {
    	this.modules[i].initialize(this.subject, this.callbackHandler, sharedState, configuration[i].getOptions());
    }

    return sharedState;
  }
  
  private void initializeAuthenticationStack() {
  	readFromFile(authenticationStack);

    if ((configuration == null) || (configuration.length == 0)) {
    	readFromFile("other");
    }    
    
    if ((configuration == null) || (configuration.length == 0)) {
      throw new IllegalStateException("No login modules configured.");
    }
  }
  
  private void readFromFile(String name) {
  	BufferedReader reader = null;
  	
    try {
    	File file = new File("LoginModuleConfiguration.txt");
    	reader = new BufferedReader(new FileReader(file));
    	String line = null;
    	
    	while ((line = reader.readLine()) != null) {
    		line = line.trim();
    		
    		if (line.startsWith(name)) {
    			int index = line.lastIndexOf("{");
    			
    			if(index > 0) {
    				line = line.substring(0, index);
    				line = line.trim();
    				
    				if (line.equals(name)) {
    					Vector<AppConfigurationEntryExtension> loginModules = new Vector<AppConfigurationEntryExtension>();
    					
    					while (((line = reader.readLine()) != null) && !line.trim().equals("};")) {
    						line = line.trim();
    						index = line.lastIndexOf(";");
    						
    						if (index > 0) {
    							line = line.substring(0, index);
    							StringTokenizer tokens = new StringTokenizer(line);
    							
    							if (tokens.countTokens() < 2) {
    								throw new IllegalArgumentException("Wrong syntax of login module configuration file, line: [" + line + "]. The line must contain at least a login module name and a flag separated with an interval.");
    							} else {
    								String loginModuleName = tokens.nextToken();
    								loginModuleName = loginModuleName.trim();
    								
    								String flag = tokens.nextToken();
    								flag = flag.trim();
    								
    								Hashtable<String, String> options = new Hashtable<String, String>();
    								
    								while (tokens.hasMoreTokens()) {
    									String option = tokens.nextToken();
    									option = option.trim();
    									
    									index = option.indexOf("=");
    									
    									if (index > 0) {
    										String optionName = option.substring(0, index);
    										optionName = optionName.trim();
    										
    										String optionValue = option.substring((index + 1), option.length());
    										optionValue = optionValue.trim();
    										
    										options.put(optionName, optionValue);
    									} else {
    										throw new IllegalArgumentException("Wrong syntax of login module configuration file, line: [" + line + "], option: [" + option + "]. The option must contain a name and a value separated by '='.");
    									}
    								}
    								
    								AppConfigurationEntryExtension loginModule = new AppConfigurationEntryExtension(loginModuleName, flag, options);
    								loginModules.add(loginModule);
    							}
    						} else {
    							throw new IllegalArgumentException("Wrong syntax at line : [" + line + "]. The line must contain a login module name and flag and must end with ';'");
    						}
    					}
    					
    					this.configuration = (AppConfigurationEntryExtension[]) loginModules.toArray(configuration);
    					
    					break;
    				}
    			}
    		}
    	}
    } catch (IOException e) {
    	LOCATION.traceThrowableT(Severity.ERROR, "Cannot read the login module configuration from the file.", e);
    	throw new IllegalStateException("Cannot read the login module configuration from the file.");
    } finally {
    	if (reader != null) {
    		try {
    			reader.close();
    		} catch (IOException e) {
    			LOCATION.traceThrowableT(Severity.ERROR, "Cannot close the login module configuration file.", e);
    		}
    	}
    }
  }
}


