package authentication.module;

import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import authentication.callback.CertificateCallback;
import authentication.credential.CertificateCredential;
import authentication.principal.CertificatePrincipal;
import authentication.principal.UserIdPrincipal;

import com.sap.engine.interfaces.security.SecurityContext;
import com.sap.engine.interfaces.security.userstore.context.UserContext;
import com.sap.engine.interfaces.security.userstore.context.UserInfo;
import com.sap.engine.services.security.server.SecurityContextImpl;
import com.sap.tc.logging.Location;


public class ClientCertificateLoginModule implements LoginModule {
  
  private static final Location LOCATION = Location.getLocation("authentication.framework");

  private CallbackHandler callbackHandler = null;
  private Subject subject = null;
  private String name = null;
  private UserContext userContext = null;
  private UserInfo user = null;
  private boolean successful;
  private boolean shouldBeIgnored;
  private SecurityContext securityContext = null;
  private X509Certificate certificate = null;

  public void initialize(Subject subject, CallbackHandler callbackHandler, Map<String, ?> sharedState, Map<String, ?> options) {
    this.callbackHandler = callbackHandler;
    this.subject = subject;

    this.successful = false;
    this.shouldBeIgnored = false;

    this.securityContext = SecurityContextImpl.getRoot();
    this.userContext = securityContext.getUserStoreContext().getActiveUserStore().getUserContext();
    if (this.userContext == null) {
      throw new SecurityException("Cannot get user context.");
    }
  }

  public boolean login() throws LoginException {
    CertificateCallback certCallback = new CertificateCallback();

    try {
      callbackHandler.handle(new Callback[] {certCallback});
    } catch (UnsupportedCallbackException e) {
      shouldBeIgnored = true;
      return false;
    } catch (IOException e) {
      throw new LoginException(e.toString());
    }

    X509Certificate[] certificates = certCallback.getCertificates();
    
    if ((certificates == null) || (certificates.length == 0) || (certificates[0] == null)) {
      if (LOCATION.beDebug()) {
        LOCATION.debugT("No certificate found.");
      }
      
      shouldBeIgnored = true;
      return false;      	
    }
     
    this.certificate = certificates[0];
    
    try {
      user = userContext.getUserInfo(certificates[0]);
    } catch (SecurityException e) {
      throw new LoginException("Wrong credentials supplied.");
    }

    if (user == null) {
      throw new LoginException("No such user " + name + " found in the userstore.");
    }

    name = user.getName();
    successful = true;
    return true;
  }

  public boolean commit() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
      	Set<UserIdPrincipal> principals = subject.getPrincipals(UserIdPrincipal.class);
      	if (principals == null || principals.isEmpty()) {
	      	CertificatePrincipal principal = new CertificatePrincipal(name);
	        subject.getPrincipals().add(principal);
      	}
        
        CertificateCredential credential = new CertificateCredential(this.certificate);
        subject.getPublicCredentials().add(credential);
        
        return true;
      } else {
        name = null;
        certificate = null;
        user = null;
        
        return false;
      }
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean abort() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
        name = null;
        certificate = null;
        successful = false;
        user = null;
      }

      return true;
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean logout() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
      	subject.getPrincipals().clear();
      	subject.getPublicCredentials().clear();
        successful = false;
      }

      return true;
    } else {
      return false;
    }
  }

}

