package authentication.module;

import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import authentication.callback.SSOTokenSetterCallback;
import authentication.credential.SSOTokenCredential;
import authentication.principal.SSOTokenPrincipal;
import authentication.principal.UserIdPrincipal;

import com.sap.engine.lib.security.Base64;
import com.sap.tc.logging.Location;


public class CreateSSOTokenLoginModule implements LoginModule {

  private static final Location LOCATION = Location.getLocation("authentication.framework");

  private CallbackHandler callbackHandler = null;
  private Subject subject = null;

  private boolean successful;
  private boolean shouldBeIgnored;

  public void initialize(Subject subject, CallbackHandler callbackHandler, Map<String, ?> sharedState, Map<String, ?> options) {
    this.callbackHandler = callbackHandler;
    this.subject = subject;

    this.successful = false;
    this.shouldBeIgnored = false;
  }

  public boolean login() throws LoginException {
    return true;
  }

  public boolean commit() throws LoginException {
  	Set<UserIdPrincipal> principals = subject.getPrincipals(UserIdPrincipal.class);

  	if (principals != null && !principals.isEmpty()) {
  		UserIdPrincipal[] array = new UserIdPrincipal[0];
  		array = (UserIdPrincipal[]) principals.toArray(array);
  		
  		UserIdPrincipal principal = array[0];
  		String name = principal.getName();
  		String cookieValue = null;
  		
  		try {
  			cookieValue = new String(Base64.encode(name.getBytes()));
  		} catch (Exception e) {
  			throw new LoginException(e.toString());
  		}
  		  		
  		boolean result = setCookie(cookieValue);
  		
  		if (result) {
  			if (principals == null || principals.isEmpty()) {
	  			SSOTokenPrincipal ssoPrincipal = new SSOTokenPrincipal(name); 
	        subject.getPrincipals().add(ssoPrincipal);
  			}
        
        SSOTokenCredential credential = new SSOTokenCredential("SSOToken", cookieValue);
	  		subject.getPrivateCredentials().add(credential);
        
	  		successful = true;
  		}
  		
  		return result;
    } else {
    	shouldBeIgnored = true;
    	return false;
    }
  }

  public boolean abort() throws LoginException {
    if (!shouldBeIgnored) {
    	boolean result = false;
    	
      if (successful) {
      	result = setCookie("");
        successful = false;
      }

      return result;
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean logout() throws LoginException {
    if (!shouldBeIgnored) {
    	boolean result = false;
    	
      if (successful) {
      	result = setCookie("");
      	subject.getPrivateCredentials().clear();
        successful = false;
      }

      return result;
    } else {
      return false;
    }
  }
  
  private boolean setCookie(String cookieValue) throws LoginException {
  	try {
  		SSOTokenSetterCallback callback = new SSOTokenSetterCallback("SSOToken", cookieValue);
	  	callbackHandler.handle(new Callback[] {callback});
    } catch (UnsupportedCallbackException e) {
      shouldBeIgnored = true;
      return false;
		} catch (Exception e) {
			throw new LoginException(e.toString());
		}
		
		return true;
  }

}
