package authentication.module;

import java.io.IOException;
import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import authentication.callback.EmailCallback;
import authentication.credential.PasswordCredential;
import authentication.principal.EmailPrincipal;
import authentication.principal.UserIdPrincipal;

import com.sap.engine.interfaces.security.SecurityContext;
import com.sap.engine.interfaces.security.userstore.context.UserContext;
import com.sap.engine.interfaces.security.userstore.context.UserInfo;
import com.sap.engine.services.security.server.SecurityContextImpl;
import com.sap.security.api.ISearchAttribute;
import com.sap.security.api.ISearchResult;
import com.sap.security.api.IUser;
import com.sap.security.api.UMFactory;
import com.sap.security.core.InternalUMFactory;
import com.sap.security.core.imp.UserSearchFilter;
import com.sap.tc.logging.Location;
import com.sap.tc.logging.Severity;


public class EmailLoginModule implements LoginModule {
  
  private static final Location LOCATION = Location.getLocation("authentication.framework");

  private CallbackHandler callbackHandler = null;
  private Subject subject = null;
  private String email = null;
  private String name = null;
  private char[] password = null;
  private UserContext userContext = null;
  private UserInfo user = null;
  private boolean successful;
  private boolean shouldBeIgnored;
  private SecurityContext securityContext = null;

  public void initialize(Subject subject, CallbackHandler callbackHandler, Map<String, ?> sharedState, Map<String, ?> options) {
    this.callbackHandler = callbackHandler;
    this.subject = subject;

    this.successful = false;
    this.shouldBeIgnored = false;
    
    this.securityContext = SecurityContextImpl.getRoot();
    this.userContext = securityContext.getUserStoreContext().getActiveUserStore().getUserContext();
    if (this.userContext == null) {
      throw new SecurityException("Cannot get user context.");
    }
  }

  public boolean login() throws LoginException {
  	EmailCallback emailCallback = new EmailCallback("Email: ");
    PasswordCallback passwordCallback = new PasswordCallback("Password: ", true);

    try {
      callbackHandler.handle(new Callback[] {emailCallback, passwordCallback});
    } catch (UnsupportedCallbackException e) {
      shouldBeIgnored = true;
      return false;
    } catch (IOException e) {
      throw new LoginException(e.toString());
    }

    email = emailCallback.getName();
    password = passwordCallback.getPassword();
    passwordCallback.clearPassword();

    if ((email == null) || (email.length() == 0)) {
      if (LOCATION.beDebug()) {
        LOCATION.debugT("No email provided.");
      }
      
      shouldBeIgnored = true;
      return false;   
    }   
     
    try {
    	UserSearchFilter searchFilter = new UserSearchFilter();
    	searchFilter.setEmail(email, ISearchAttribute.EQUALS_OPERATOR, false);
    	ISearchResult searchResult = UMFactory.getUserFactory().searchUsers(searchFilter);
    	
    	if ((searchResult != null) && searchResult.hasNext()) {
    		name = (String) searchResult.next();
    		LOCATION.infoT("user unique id: " + name);
    		
    		if (name != null) {
    			name = name.substring(name.lastIndexOf(":") + 1);
    			user = userContext.getUserInfo(name);
    			LOCATION.infoT("user: " + user);
    		}
    	}
    } catch (Exception e) {
    	LOCATION.traceThrowableT(Severity.ERROR, "Error: ", e);
      throw new LoginException("Wrong username / password combination.");
    }

    if (user == null) {
      throw new LoginException("No such user with email " + email + " found in the userstore.");
    }

    int passwordStatus = user.checkPasswordExtended(password);

    if (passwordStatus == UserInfo.CHECKPWD_WRONGPWD) {
    	throw new LoginException("Invalid password for user " + name + " .");
    }

    name = user.getName();
    LOCATION.infoT("user name " + name);
    successful = true;
    return true;
  }

  public boolean commit() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
      	Set<UserIdPrincipal> principals = subject.getPrincipals(UserIdPrincipal.class);
      	if (principals == null || principals.isEmpty()) {
	      	EmailPrincipal principal = new EmailPrincipal(name);   
	        subject.getPrincipals().add(principal);
	        LOCATION.errorT("Email principal added.");
      	}
        
        subject.getPrivateCredentials().add(new PasswordCredential(name, password));
        
        return true;
      } else {
        email = null;
        name = null;
        password = null;
        user = null;
        
        return false;
      }
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean abort() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
        email = null;
        name = null;
        password = null;
        successful = false;
        user = null;
      }

      return true;
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean logout() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
      	subject.getPrincipals().clear();
      	subject.getPrivateCredentials().clear();
        successful = false;
      }

      return true;
    } else {
      return false;
    }
  }

}
