package authentication.module;

import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import authentication.callback.SSOTokenGetterCallback;
import authentication.credential.SSOTokenCredential;
import authentication.principal.SSOTokenPrincipal;
import authentication.principal.UserIdPrincipal;

import com.sap.engine.interfaces.security.SecurityContext;
import com.sap.engine.interfaces.security.userstore.context.UserContext;
import com.sap.engine.interfaces.security.userstore.context.UserInfo;
import com.sap.engine.lib.security.Base64;
import com.sap.engine.services.security.server.SecurityContextImpl;
import com.sap.tc.logging.Location;


public class EvaluateSSOTokenLoginModule implements LoginModule {

  private static final Location LOCATION = Location.getLocation("authentication.framework");

  private CallbackHandler callbackHandler = null;
  private Subject subject = null;

  private boolean successful;
  private boolean shouldBeIgnored;
  
  private UserInfo user = null;
  private UserContext userContext = null;
  private SecurityContext securityContext = null;
  
  private String name = null;
  private String cookieValue = null;

  public void initialize(Subject subject, CallbackHandler callbackHandler, Map<String, ?> sharedState, Map<String, ?> options) {
    this.callbackHandler = callbackHandler;
    this.subject = subject;

    this.successful = false;
    this.shouldBeIgnored = false;
    
    this.securityContext = SecurityContextImpl.getRoot();
    this.userContext = securityContext.getUserStoreContext().getActiveUserStore().getUserContext();
    if (this.userContext == null) {
      throw new SecurityException("Cannot get user context.");
    }
  }

  public boolean login() throws LoginException {
  	SSOTokenGetterCallback callback = new SSOTokenGetterCallback("SSOToken");
  	
  	try {
	  	this.callbackHandler.handle(new Callback[] {callback});
	  	this.cookieValue = callback.getValue();
	  	
	  	if (this.cookieValue != null) {
	  		this.name = new String(Base64.decode(cookieValue.getBytes()));
	  	} else {
	  		shouldBeIgnored = true;
	  	}
    } catch (UnsupportedCallbackException e) {
      shouldBeIgnored = true;
      return false;
    } catch (Exception e) {
      throw new LoginException(e.toString());
    }
		
		if (name != null) {
			try {
				user = userContext.getUserInfo(name);
	    } catch (SecurityException e) {
	      throw new LoginException("Wrong username / password combination.");
	    }
		}

    if (user == null) {
      throw new LoginException("No such user " + name + " found in the userstore.");
    }
  	
    successful = true;
    return true;
  }

  public boolean commit() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
      	Set<UserIdPrincipal> principals = subject.getPrincipals(UserIdPrincipal.class);
      	if (principals == null || principals.isEmpty()) {
	      	SSOTokenPrincipal principal = new SSOTokenPrincipal(name);   
	        subject.getPrincipals().add(principal);
      	}
      	
        subject.getPrivateCredentials().add(new SSOTokenCredential("SSOToken", cookieValue));
        
        return true;
      } else {
        name = null;
        user = null;
        
        return false;
      }
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean abort() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
        name = null;
        successful = false;
        user = null;
      }

      return true;
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean logout() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
      	subject.getPrincipals().clear();
      	subject.getPrivateCredentials().clear();
        successful = false;
      }

      return true;
    } else {
      return false;
    }
  }

}
