package authentication.module;

import java.io.IOException;
import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import authentication.credential.PasswordCredential;
import authentication.principal.NameAndPasswordPrincipal;
import authentication.principal.UserIdPrincipal;

import com.sap.engine.interfaces.security.SecurityContext;
import com.sap.engine.interfaces.security.userstore.context.UserContext;
import com.sap.engine.interfaces.security.userstore.context.UserInfo;
import com.sap.engine.services.security.server.SecurityContextImpl;
import com.sap.tc.logging.Location;


public class NameAndPasswordLoginModule implements LoginModule {
  
  private static final Location LOCATION = Location.getLocation("authentication.framework");

  private CallbackHandler callbackHandler = null;
  private Subject subject = null;
  private String name = null;
  private char[] password = null;
  private UserContext userContext = null;
  private UserInfo user = null;
  private boolean successful;
  private boolean shouldBeIgnored;
  private SecurityContext securityContext = null;

  public void initialize(Subject subject, CallbackHandler callbackHandler, Map<String, ?> sharedState, Map<String, ?> options) {
    this.callbackHandler = callbackHandler;
    this.subject = subject;

    this.successful = false;
    this.shouldBeIgnored = false;

    this.securityContext = SecurityContextImpl.getRoot();
    this.userContext = securityContext.getUserStoreContext().getActiveUserStore().getUserContext();
    if (this.userContext == null) {
      throw new SecurityException("Cannot get user context.");
    }
  }

  public boolean login() throws LoginException {
    NameCallback nameCallback = new NameCallback("User name: ");
    PasswordCallback passwordCallback = new PasswordCallback("Password: ", true);

    try {
      callbackHandler.handle(new Callback[] {nameCallback, passwordCallback});
    } catch (UnsupportedCallbackException e) {
      shouldBeIgnored = true;
      return false;
    } catch (IOException e) {
      throw new LoginException(e.toString());
    }

    name = nameCallback.getName();
    password = passwordCallback.getPassword();
    passwordCallback.clearPassword();

    if ((name == null) || (name.length() == 0)) {
      if (LOCATION.beDebug()) {
        LOCATION.debugT("No user name provided.");
      }
      
      shouldBeIgnored = true;
      return false;   
    }   
     
    try {
      user = userContext.getUserInfo(name);
      LOCATION.infoT("user: " + user);
    } catch (SecurityException e) {
      throw new LoginException("Wrong username / password combination.");
    }

    if (user == null) {
      throw new LoginException("No such user " + name + " found in the userstore.");
    }

    int passwordStatus = user.checkPasswordExtended(password);

    if (passwordStatus == UserInfo.CHECKPWD_WRONGPWD) {
    	throw new LoginException("Invalid password for user " + name + " .");
    }
    LOCATION.infoT("user name " + name);
    name = user.getName();
    successful = true;
    return true;
  }

  public boolean commit() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
      	Set<UserIdPrincipal> principals = subject.getPrincipals(UserIdPrincipal.class);

      	if (principals == null || principals.isEmpty()) {
      		NameAndPasswordPrincipal principal = new NameAndPasswordPrincipal(name);
      		subject.getPrincipals().add(principal);
      		LOCATION.infoT("Name principal set.");
      	}
        
        subject.getPrivateCredentials().add(new PasswordCredential(name, password));
      } else {
        name = null;
        password = null;
        user = null;
      }

      return true;
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean abort() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
        name = null;
        password = null;
        successful = false;
        user = null;
      }

      return true;
    } else {
      shouldBeIgnored = false;
      return false;
    }
  }

  public boolean logout() throws LoginException {
    if (!shouldBeIgnored) {
      if (successful) {
      	subject.getPrincipals().clear();
      	subject.getPrivateCredentials().clear();
        successful = false;
      }

      return true;
    } else {
      return false;
    }
  }

}
